//  -*- Mode: Java -*-
//
// /tmp/build-stella/stella-3.0.2.beta/native/java/edu/isi/stella/Set.java

/*--------------------------- BEGIN LICENSE BLOCK ---------------------------+
|                                                                            |
| Version: MPL 1.1/GPL 2.0/LGPL 2.1                                          |
|                                                                            |
| The contents of this file are subject to the Mozilla Public License        |
| Version 1.1 (the "License"); you may not use this file except in           |
| compliance with the License. You may obtain a copy of the License at       |
| http://www.mozilla.org/MPL/                                                |
|                                                                            |
| Software distributed under the License is distributed on an "AS IS" basis, |
| WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License   |
| for the specific language governing rights and limitations under the       |
| License.                                                                   |
|                                                                            |
| The Original Code is the STELLA Programming Language.                      |
|                                                                            |
| The Initial Developer of the Original Code is                              |
| UNIVERSITY OF SOUTHERN CALIFORNIA, INFORMATION SCIENCES INSTITUTE          |
| 4676 Admiralty Way, Marina Del Rey, California 90292, U.S.A.               |
|                                                                            |
| Portions created by the Initial Developer are Copyright (C) 1996-2001      |
| the Initial Developer. All Rights Reserved.                                |
|                                                                            |
| Contributor(s):                                                            |
|                                                                            |
| Alternatively, the contents of this file may be used under the terms of    |
| either the GNU General Public License Version 2 or later (the "GPL"), or   |
| the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),   |
| in which case the provisions of the GPL or the LGPL are applicable instead |
| of those above. If you wish to allow use of your version of this file only |
| under the terms of either the GPL or the LGPL, and not to allow others to  |
| use your version of this file under the terms of the MPL, indicate your    |
| decision by deleting the provisions above and replace them with the notice |
| and other provisions required by the GPL or the LGPL. If you do not delete |
| the provisions above, a recipient may use your version of this file under  |
| the terms of any one of the MPL, the GPL or the LGPL.                      |
|                                                                            |
+---------------------------- END LICENSE BLOCK ----------------------------*/

package edu.isi.stella;

import edu.isi.stella.javalib.*;

public class Set extends Collection {
    public Cons theConsList;

  public static Set newSet() {
    { Set self = null;

      self = new Set();
      self.theConsList = Stella.NIL;
      return (self);
    }
  }

  public AbstractIterator allocateIterator() {
    { Set self = this;

      { SetIterator iterator = SetIterator.newSetIterator();

        iterator.setIteratorCursor = self.theConsList;
        iterator.setIteratorCollection = self;
        iterator.firstIterationP = Stella.TRUE;
        return (iterator);
      }
    }
  }

  /** Make <code>self</code> an empty set.
   */
  public void clear() {
    { Set self = this;

      self.theConsList = Stella.NIL;
    }
  }

  /** Return a copy of the set <code>self</code>.  The conses in the copy are
   * freshly allocated.
   * @return Set
   */
  public Set copy() {
    { Set self = this;

      { Set copy = ((Set)(Surrogate.createObject(self.primaryType(), Stella.NIL)));

        copy.theConsList = Cons.copyConsList(self.theConsList);
        return (copy);
      }
    }
  }

  public Set reverse() {
    { Set self = this;

      self.theConsList = self.theConsList.reverse();
      return (self);
    }
  }

  /** Remove all entries in <code>self</code> that match <code>value</code>.
   * @param value
   * @return AbstractCollection
   */
  public AbstractCollection remove(Stella_Object value) {
    { Set self = this;

      self.theConsList = self.theConsList.remove(value);
      return (self);
    }
  }

  /** Add <code>value</code> to the set <code>self</code>.  First checks for duplicate.
   * @param value
   */
  public void insert(Stella_Object value) {
    { Set self = this;

      if (!self.memberP(value)) {
        self.theConsList = Stella_Object.cons(value, self.theConsList);
      }
    }
  }

  public int length() {
    { Set self = this;

      return (self.theConsList.length());
    }
  }

  /** Return <code>true</code> iff <code>renamed_Object</code> is a member of the set <code>self</code>.
   * Uses an <code>eqlP</code> test.
   * @param renamed_Object
   * @return boolean
   */
  public boolean memberP(Stella_Object renamed_Object) {
    { Set self = this;

      { Stella_Object i = null;
        Cons iter000 = self.theConsList;

        while (!(iter000 == Stella.NIL)) {
          {
            i = iter000.value;
            iter000 = iter000.rest;
          }
          if (Stella_Object.eqlP(i, renamed_Object)) {
            return (Stella.TRUE);
          }
        }
      }
      return (Stella.FALSE);
    }
  }

  public static boolean terminateSetP(Set self) {
    { Cons cons = self.theConsList;
      Cons rest = null;

      while (!(cons == Stella.NIL)) {
        rest = cons.rest;
        cons.free();
        cons = rest;
      }
      self.theConsList = Stella.NIL;
      return (Stella.TRUE);
    }
  }

  /** Return <code>true</code> if the set <code>self</code> has at least one member.
   * @return boolean
   */
  public boolean nonEmptyP() {
    { Set self = this;

      return (!(self.theConsList == Stella.NIL));
    }
  }

  /** Return <code>true</code> if the set <code>self</code> has no members.
   * @return boolean
   */
  public boolean emptyP() {
    { Set self = this;

      return (self.theConsList == Stella.NIL);
    }
  }

  public boolean noDuplicatesP() {
    { Set self = this;

      return (Stella.TRUE);
    }
  }

  public void printObject(java.io.PrintStream stream) {
    { Set self = this;

      Cons.printConsList(self.theConsList, stream);
    }
  }

  public static Stella_Object accessSetSlotValue(Set self, Symbol slotname, Stella_Object value, boolean setvalueP) {
    if (slotname == Stella.SYM_STELLA_THE_CONS_LIST) {
      if (setvalueP) {
        self.theConsList = ((Cons)(value));
      }
      else {
        value = self.theConsList;
      }
    }
    else {
      { OutputStringStream stream000 = OutputStringStream.newOutputStringStream();

        stream000.nativeStream.print("`" + slotname + "'" + " is not a valid case option");
        throw StellaException.newStellaException(stream000.theStringReader());
      }
    }
    return (value);
  }

  public Surrogate primaryType() {
    { Set self = this;

      return (Stella.SGT_STELLA_SET);
    }
  }

}

